#include <stdio.h>
#include <string.h>
#include <sys/stat.h>

long i;

FILE* infile;
FILE* outfile;

char datablock[32768];


char ChunkID[5],Format[5],Subchunk1ID[5],Subchunk2ID[5];
char SubChunkID[5];
char infilename[256];
char outfilename[256];
char settingsfilename[256];
char appdirname[256];
char prefix[128];
char outdirname[256];
char leafname[128];

long ChunkSize,Subchunk1Size,AudioFormat;
long NumChannels,SampleRate,ByteRate,BlockAlign;
long BitsPerSample,BytesPerSample;

char BWAVEDescription[256];
char BWAVEOriginator[32];
char BWAVEOriginatorRef[32];
char BWAVEOriginationDate[10];
char BWAVEOriginationTime[8];
char UMID[64];
char BWAVEReserved[190];

long TimeRefLow, TimeRefHigh, BWAVEVersion,bhcount,chcount;
long BWAVEChunkSize,again,GeneralChunkSize,payload;
long again,bhcount;

long icut,icut2,slat,file_size;
long bwave_found,general_found,data_found,format_found,header_size;
long blocksize,size_read,ok_to_clean;

unsigned long Subchunk2Size;

long read_header(void);
void read_settings(void);
long read_input_file_size(char*);
long scan_header(void);
long read_data_chunk(void);
long read_general_chunk(void);
long read_format_chunk(void);
long read_bwave_chunk(void);

void get4(char*);
long get4int(void);
long get3int(void);
long get2int(void);

void write4int(long);
void write2int(long);
void write4(char*);
void write_output_header(long);


int main(int argc,char*argv[])
{
  icut=32768; icut2=2*icut;
  blocksize=32768;
  
  printf("WAVCleaner v 1.00 (22 May 2010)\n");
  printf("Examines an LPCM WAV file and if it has an\n");
  printf("extended header generates a cleaned new version.\n\n");
  
  i=argc;
  printf("Number of arguments given = %ld\n\n",i);

  if(i>0) printf("Application at  = %s\n",argv[0]);
  
  sprintf(outfilename,"%s",argv[0]);
  
  slat=(long)strrchr(outfilename,'/')-(long)outfilename;
  outfilename[slat]=(char)0;
  sprintf(appdirname,"%s",outfilename);
  sprintf(settingsfilename,"%s/Settings.txt\0",appdirname);
  
  
  printf("WAVCleaner dir at %s\n",appdirname);
  printf("Settings file at %s\n",settingsfilename);
  
  read_settings();
  
  
  
  if(i>1)
  {
    printf("Input File at = %s\n",argv[1]);

    sprintf(infilename,"%s\0",argv[1]);
    
    sprintf(leafname,"%s\0",1+strrchr(infilename,'/'));
    printf("leafname = %s\n",leafname);
    
    sprintf(outfilename,"%s/%s%s\0",outdirname,prefix,leafname);
    printf("Output to file\n  %s\n",outfilename);
    

    file_size=read_input_file_size(infilename);
    
    printf("Input file size = %ld bytes\n",file_size);
    
    infile=fopen(infilename,"rd");
    
    
    scan_header();
    
    printf("Chunk types found [1 = yes : 0 = no]\n");
    printf("   format = %ld",format_found);
    printf("   bwave  = %ld",bwave_found);
    printf(" extended = %ld",general_found);
    printf("   data   = %ld \n",data_found);
    header_size=file_size-payload;
    printf("Header size = %ld Bytes\n",header_size);
    
    ok_to_clean=1;
    if((format_found*data_found)==1)
    {
      if ((bwave_found==0)&&(general_found==0))
      {
        if (header_size==44)
        {
          printf("Header already clean OK!\n");
          ok_to_clean=0;
    
        }
      }
    }
    
    if (ok_to_clean==1)
    {
      printf("Making clean copy\n");
      outfile=fopen(outfilename,"wd");
      write_output_header(payload);
      
      do
      {
        
        size_read=fread(datablock,sizeof(char),blocksize,infile);
        fwrite(datablock,sizeof(char),size_read,outfile);
      
      } while (size_read==blocksize);
      
      fclose(outfile);
    }
    
    
    fclose(infile);
  
  }
  else
  {
    printf("No input file!\n");
    printf("Drag a WAV file to the WAVCleaner application icon\n");
    printf("to see the results\n");
  }
       
  printf("Done!\n");
  return 0;
}

long scan_header(void)
{
  long headtype;
  headtype=0; bhcount=0; again=0;
  header_size=0;
  bwave_found=0; general_found=0; data_found=0; format_found=0;

  get4(ChunkID);
  printf("\nChunkID = %s\n",ChunkID);
  ChunkSize=get4int();
  printf("Chunk Size = %ld\n",ChunkSize);
  get4(Format);
  printf("Format = %s\n\n\n",Format);
  
  if (strcmp(Format,"WAVE")==0)
  {
    printf("WAVE found\n");
    
    do
    {
      
      get4(SubChunkID);
      printf("\n\nSubchunkID = %s\n",SubChunkID);
      again=0;
      
      if(strcmp(SubChunkID,"bext\0")==0)
      {
        printf("\nBWAVE Chunk found\n");
        bwave_found=1;
        again=1;
        read_bwave_chunk();
      }
      
      if(strcmp(SubChunkID,"fmt \0")==0)
      {
        printf("Format chunk found\n");
        format_found=1;
        again=1;
        read_format_chunk();
      }
      
      
      
      if(again==0)
      {
        if(strcmp(SubChunkID,"data\0")==0)
        {
          printf("Data chunk found\n");
          data_found=1;
          read_data_chunk();
        }
        else
        {
          printf("General chunk found\n");
          general_found=1;
          read_general_chunk();
          again=1;
        }
      }
       
      
    
    } while (again==1);
  }
  
  return headtype;
}


long read_data_chunk(void)
{
  Subchunk2Size=get4int();
  printf("Subchunk2Size = %ld [Number of Data payload bytes]\n",Subchunk2Size);
  payload=Subchunk2Size; 
  return 1;
}


long read_general_chunk(void)
{
  char c1;
  GeneralChunkSize=get4int();
  printf("Size = %ld\n",GeneralChunkSize);
  bhcount=0;
  do
  {
    fscanf(infile,"%c",&c1);
    if ((int)c1!=0) printf("%c",c1);
    bhcount++;
  } while (bhcount<GeneralChunkSize);
  printf("\n");
  return 1;
}


long read_format_chunk(void)
{
   char c1;
    Subchunk1Size=get4int();
    printf(" Sub chunk size = %ld\n",Subchunk1Size);
    bhcount=0;
    
    AudioFormat=get2int(); bhcount+=2;
    
    if(AudioFormat==1)
    {
      printf("16 bits per sample LPCM format OK\n");
    }
    else
    {
      printf("Format unknown!\n");
    }
    
    NumChannels=get2int(); bhcount+=2;
    printf("Number of Channels = %ld\n",NumChannels);
    SampleRate=get4int(); bhcount+=4;
    printf("Sample Rate = %ld\n",SampleRate);
    ByteRate=get4int(); bhcount+=4;
    printf("Byte Rate = %ld\n",ByteRate);
    
    BlockAlign=get2int(); bhcount+=2;
    printf("Block Align = %ld\n",BlockAlign);
    BitsPerSample=get2int(); bhcount+=2;
    printf("Bits per sample = %ld\n",BitsPerSample);
    BytesPerSample=BitsPerSample/8;
    printf("Bytes per sample = %ld\n",BytesPerSample);
    
    while(bhcount<Subchunk1Size)
    {
      bhcount++;
      fscanf(infile,"%c",&c1);
    }
    
    printf("bhcount %ld\n",bhcount);
  
  return 1;
}

long read_bwave_chunk(void)
{
  char c1;
  printf("Reading Broadcast Wav header chunk\n");
    
    BWAVEChunkSize=get4int();
    
    printf("Size = %ld\n",BWAVEChunkSize);
    
    bhcount=0;
   
    i=0;
    do
    {
      fscanf(infile,"%c",&c1);
      BWAVEDescription[i]=c1;
      i++; bhcount++;
    } while (i<256);
    
    printf("%3ld Description => %s\n",bhcount,BWAVEDescription);
    
    i=0;
    do
    {
      fscanf(infile,"%c",&c1);
      BWAVEOriginator[i]=c1;
      i++; bhcount++;
    } while (i<32);
    
    printf("%3ld Originator => %s\n",bhcount,BWAVEOriginator);
    
    
    i=0;
    do
    {
      fscanf(infile,"%c",&c1);
      BWAVEOriginatorRef[i]=c1;
      i++; bhcount++;
    } while (i<32);
    
    printf("%3ld Reference => %s\n",bhcount,BWAVEOriginatorRef);
    
    i=0;
    do
    {
      fscanf(infile,"%c",&c1);
      BWAVEOriginationDate[i]=c1;
      i++; bhcount++;
    } while (i<10);
    
    printf("%3ld Date => %s\n",bhcount,BWAVEOriginationDate);
    
    
    i=0;
    do
    {
      fscanf(infile,"%c",&c1);
      BWAVEOriginationTime[i]=c1;
      i++; bhcount++;
    } while (i<8);
    
    printf("%3ld Time => %s\n",bhcount,BWAVEOriginationTime);
    
    TimeRefLow=get4int(); bhcount+=4;
    TimeRefHigh=get4int(); bhcount+=4;
    
    printf("%3ld Time ref values  %ld  %ld\n",bhcount,TimeRefHigh,TimeRefLow);
    
    BWAVEVersion=get2int(); bhcount+=2;
    
    printf("%3ld BWAVE Version = %ld\n",bhcount,BWAVEVersion);
    
    i=0;
    do
    {
      fscanf(infile,"%c",&c1);
      UMID[i]=c1;
      
      if((int)c1>31) printf("%ld %c\n",bhcount,c1);
      
      i++; bhcount++;
    } while (i<64);
    
    printf("%3ld Read UMID\n",bhcount);
    
    i=0;
    do
    {
      fscanf(infile,"%c",&c1);
      BWAVEReserved[i]=c1;
      
      if((int)c1>31) printf("%ld %c\n",i,c1);
      
      i++; bhcount++;
    } while (i<190);
    
    printf("%3ld Read BWAVE Reserved\n",bhcount);
    
    i=0;
    do
    {
      fscanf(infile,"%c",&c1);
      if((int)c1!=0) printf("%ld %d %c\n",bhcount,(int)c1,c1);
      i++; bhcount++;
    } while (bhcount<BWAVEChunkSize);
    
    printf("Reached end of BWAVE chunk header\n");
  
  return 1;
}

long read_input_file_size(char* filename)
{
  struct stat fileinfo;
  long result;
  result=-1;
  if (stat(filename,&fileinfo)==0)
  {
    result=(long)fileinfo.st_size;
    printf("file %s is\n   %ld bytes long\n",filename,result);
  }
  else
  {
    printf("file %s not found!\n",filename);
  }
  
  return result;
}

void read_settings(void)
{
  infile=fopen(settingsfilename,"rt");
  
  fscanf(infile,"%s",prefix);
  printf("Output prefix = %s\n",prefix);
  fscanf(infile,"%s",outdirname);
  printf("Output to dir = %s\n",outdirname);
  
  
  fclose(infile);
}


long read_header(void)
{
  long irh;

   printf("Reading header\n");
   
   ByteRate=0;
    
  irh=ByteRate;
  return irh;
}


long get2int(void)
{
  long result;
  unsigned char cip1,cip2;
  
  fscanf(infile,"%c",&cip1);
  fscanf(infile,"%c",&cip2);

  result=(long)cip2;
  result=(long)cip1+256*result;

  return result;
} 

long get4int(void)
{
  long result;
  unsigned char cip1,cip2,cip3,cip4;
  
  fscanf(infile,"%c",&cip1);
  fscanf(infile,"%c",&cip2);
  fscanf(infile,"%c",&cip3);
  fscanf(infile,"%c",&cip4);
    
  result=(long)cip4;
  result=(long)cip3+256*result;
  result=(long)cip2+256*result;
  result=(long)cip1+256*result;
  
  return result;
} 
void get4(char* outstring)
{
  long isp;
  char csp;
  isp=0;
  do
  {
    fscanf(infile,"%c",&csp);
    outstring[isp]=csp;
    isp++;
  } while (isp<4);
  outstring[4]=(char)0;

}


void write4int(long inv)
{
  char cop1,cop2,cop3,cop4;
  
  cop1=(unsigned char)(inv&255);
  cop2=(unsigned char)((inv>>8)&255);
  cop3=(unsigned char)((inv>>16)&255);
  cop4=(unsigned char)((inv>>24)&255);
  
  fprintf(outfile,"%c",cop1);
  fprintf(outfile,"%c",cop2);
  fprintf(outfile,"%c",cop3);
  fprintf(outfile,"%c",cop4);
}


void write2int(long inv)
{
  unsigned char cop1,cop2;
  cop1=(unsigned char)(inv&255);
  cop2=(unsigned char)((inv>>8)&255);
  fprintf(outfile,"%c",cop1);
  fprintf(outfile,"%c",cop2);
}

void write4(char* outstring)
{
  int isp;
  
  isp=0;
  do
  {
    fprintf(outfile,"%c",outstring[isp]);
    isp++;
  } while (isp<4);
  
}


void write_output_header(long plo)
{
  
  printf("Payload to write = %ld\n\n",plo);
  write4("RIFF\0");
  write4int(plo+36);
  write4("WAVE\0");
  write4("fmt \0");
  write4int(16);
  write2int(AudioFormat);
  write2int(NumChannels);
  write4int(SampleRate);
  write4int(ByteRate);
  write2int(BlockAlign);
  write2int(BitsPerSample);
  write4("data\0");
  write4int(plo);
  
}








